// file      : xsd/cxx/tree/counter.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/counter.hxx>

#include <iostream>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      //
      //
      struct GlobalType: Traversal::Type
      {
        GlobalType (UnsignedLong& count)
            : count_ (count)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type&)
        {
          ++count_;
        }

      private:
        UnsignedLong& count_;
      };

      //
      //
      struct GlobalElement: Traversal::Element,
                            GlobalElementBase,
                            Context
      {
        GlobalElement (Context& c,
                       UnsignedLong& count,
                       UnsignedLong& gen_count)
            : GlobalElementBase (c),
              Context (c),
              count_ (count), gen_count_ (gen_count), last_ (0)
        {
        }

        ~GlobalElement ()
        {
          if (last_ != 0)
          {
            last_->context ().set ("last", true);

            if (generate_p (*last_))
              ++gen_count_;
          }
        }

        virtual Void
        traverse (Type& e)
        {
          // Check if the previous element we saw needs to be generated.
          //
          if (last_ != 0 && generate_p (*last_))
            ++gen_count_;

          last_ = &e;

          if (count_ == 0)
            e.context ().set ("first", true);

          ++count_;
        }

      private:
        SemanticGraph::Element* last_;
        UnsignedLong& count_;
        UnsignedLong& gen_count_;
      };
    }

    Counter::
    Counter ()
    {
    }

    Counts Counter::
    count (CLI::Options const& options, SemanticGraph::Schema& tu)
    {
      Counts counts;
      Context ctx (std::wcerr, tu, options, counts);

      Traversal::Schema schema;
      Traversal::Sources sources;

      schema >> sources >> schema;

      Traversal::Names schema_names;
      Traversal::Namespace ns;
      Traversal::Names ns_names;
      GlobalType global_type (counts.global_types);
      GlobalElement global_element (
        ctx, counts.global_elements, counts.generated_global_elements);

      schema >> schema_names >> ns >> ns_names;

      ns_names >> global_element;
      ns_names >> global_type;

      schema.dispatch (tu);

      return counts;
    }
  }
}
