// file      : xsd/cxx/tree/serialization-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/serialization-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c,
              String const& name_ = "",
              String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
          {
            name = ename (l);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }


          // operator<< (xercesc::DOMElement)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMElement& e," << endl
             << "const " << type_name << "& v)"
             << "{"
             << "e << static_cast< const ::xsd::cxx::tree::list< " <<
            item_type_name (l.argumented ().type ()) << ", " <<
            char_type << " >& > (v);"
             << "}";

          // operator<< (xercesc::DOMAttr)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMAttr& a," << endl
             << "const " << type_name << "& v)"
             << "{"
             << "a << static_cast< const ::xsd::cxx::tree::list< " <<
            item_type_name (l.argumented ().type ()) << ", "
             << char_type << " >& > (v);"
             << "}";

          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " << char_type
             << " >& l," << endl
             << "const " << type_name << "& v)"
             << "{"
             << "l << static_cast< const ::xsd::cxx::tree::list< " <<
            item_type_name (l.argumented ().type ()) << ", " <<
            char_type << " >& > (v);"
             << "}";

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () && l.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " << type_name << " >" << endl
               << "_xsd_" << name << "_type_name_init (" << endl
               << L << "\"" << l.name () << "\"," << endl
               << L << "\"" << xml_ns_name (l) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String
        item_type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

      private:
        String name;
        String type_name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c,
               String const& name_ = "",
               String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
          {
            name = ename (u);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          String base (xs_ns_name () + L"::string");

          // operator<< (xercesc::DOMElement)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMElement& e," << endl
             << "const " << type_name << "& v)"
             << "{"
             << "e << static_cast< const " << base << "& > (v);"
             << "}";

          // operator<< (xercesc::DOMAttr)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMAttr& a," << endl
             << "const " << type_name << "& v)"
             << "{"
             << "a << static_cast< const " << base << "& > (v);"
             << "}";

          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " << char_type
             << " >& l," << endl
             << "const " << type_name << "& v)"
             << "{"
             << "l << static_cast< const " << base << "& > (v);"
             << "}";

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () && u.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " << type_name << " >" << endl
               << "_xsd_" << name << "_type_name_init (" << endl
               << L << "\"" << u.name () << "\"," << endl
               << L << "\"" << xml_ns_name (u) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String name;
        String type_name;
      };


      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c,
                     String const& name_ = "",
                     String const& type_name_ = "")
            : Context (c),
              base_ (c),
              name (name_),
              type_name (type_name_)
        {
          inherits_base_ >> base_;
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
          {
            name = ename (e);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          // operator<< (xercesc::DOMElement)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMElement& e," << endl
             << type_name << " i)"
             << "{"
             << "e << static_cast< const ";

          inherits (e, inherits_base_);

          os << "& > (i);"
             << "}";


          // operator<< (xercesc::DOMAttr)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMAttr& a," << endl
             << type_name << " i)"
             << "{"
             << "a << static_cast< const ";

          inherits (e, inherits_base_);

          os << "& > (i);"
             << "}";


          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " << char_type
             << " >& l," << endl
             << type_name << " i)"
             << "{"
             << "l << static_cast< const ";

          inherits (e, inherits_base_);

          os << "& > (i);"
             << "}";


          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () && e.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " << type_name << " >" << endl
               << "_xsd_" << name << "_type_name_init (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << xml_ns_name (e) << "\");"
               << endl
               << endl;
          }
        }

      private:
        Traversal::Inherits inherits_base_;
        BaseTypeName base_;

        String name;
        String type_name;
      };


      //
      //
      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            // I need to build scope information with the same names
            // as in the header so that I get member escaping right.
            // Actual names will be from the public interface.
            //

            String const& name (ename (m));
            String type_name (scope
                              + L"::" + name
                              + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, name, type_name);
            Union union_ (*this, name, type_name);
            Complex complex (*this, name, type_name);
            Enumeration enumeration (*this, name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }

      private:
        String scope;
      };


      struct Element : Traversal::Element, protected virtual Context
      {
        Element (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (skip (e)) return;

          String const& name (ename (e));
          String ns (e.qualified () ? e.namespace_ ().name () : "");
          String type (scope
                       + L"::" + name
                       + L"::" + unclash (name, "type"));

          // Check if we need to handle xsi:type and substitution groups.
          // If this elemens type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          SemanticGraph::Type& t (e.type ());
          Boolean polymorphic (
            options.value<CLI::generate_polymorphic> () && t.named ());

          if (polymorphic)
          {
            Boolean fund (false);
            IsFundamentalType traverser (fund);
            traverser.dispatch (t);
            polymorphic = !fund;
          }

          os << "{";

          // aCC cannot handle an inline call to type_name_map_instance.
          //
          if (polymorphic)
            os << "::xsd::cxx::tree::type_name_map< " << char_type
               << " >& tnm (" << endl
               << "::xsd::cxx::tree::type_name_map_instance< 0, " <<
              char_type << " > ());"
               << endl;

          if (max (e) != 1)
          {
            // sequence
            //
            os << "for (" << scope << "::" << name << "::" <<
              unclash (name, "const_iterator") << endl
               << "b (i." << name << " ().begin ()), " <<
              "n (i." << name << " ().end ());" << endl
               << "b != n; ++b)"
               << "{";

            if (polymorphic)
            {
              os << "tnm.serialize< " << type << " > (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << "," << endl
                 << (e.qualified () ? "true" : "false") << "," << endl
                 << "e," << endl
                 << "*b);";
            }
            else
            {
              os << xml_element_type << " s (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e);"
                 << "s.dom_element () << *b;";
            }

            os << "}";
          }
          else if (min (e) == 0)
          {
            // optional
            //
            os << "if (i." << name << " ())"
               << "{";

            if (polymorphic)
            {
              os << "tnm.serialize< " << type << " > (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << "," << endl
                 << (e.qualified () ? "true" : "false") << "," << endl
                 << "e," << endl
                 << "*i." << name << " ());";
            }
            else
            {
              os << xml_element_type << " s (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e);"
                 << "s.dom_element () << *i." << name << " ();";
            }

            os << "}";
          }
          else
          {
            // one
            //
            if (polymorphic)
            {
              os << "tnm.serialize< " << type << " > (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << "," << endl
                 << (e.qualified () ? "true" : "false") << "," << endl
                 << "e," << endl
                 << "i." << name << " ());";
            }
            else
            {
              os << xml_element_type << " s (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e);"
                 << "s.dom_element () << i." << name << " ();";
            }
          }

          os << "}";
        }

      private:
        String scope;
      };

      struct Attribute : Traversal::Attribute, protected virtual Context
      {
        Attribute (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));
          String ns (a.qualified () ? a.namespace_ ().name () : "");

          if (a.optional () && !a.default_ ())
          {
            os << "if (i." << name << " ())"
               << "{"
               << xml_attribute_type << " a (" << endl
               << L << "\"" << a.name () << "\"," << endl
               << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
               << "e);"
               << endl
               << "a.dom_attribute () << *i." << name << " ();"
               << "}";
          }
          else
          {
            os << "{"
               << xml_attribute_type << " a (" << endl
               << L << "\"" << a.name () << L"\"," << endl
               << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
               << "e);"
               << endl
               << "a.dom_attribute () << i." << name << " ();"
               << "}";
          }
        }
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c,
                 String const& name_ = "",
                 String const& type_name_ = L"")
            : Context (c),
              name (name_),
              type_name (type_name_),
              base_ (c)
        {
          inherits_ >> base_;
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
          {
            name = ename (c);

            enter_scope (name);
            type_name = scope;
            leave_scope ();
          }

          // Go after anonymous types first.
          //
          {
            enter_scope (name);

            Member<Complex> member (*this, type_name);
            Traversal::Names names_member (member);

            names (c, names_member);

            leave_scope ();
          }


          // operator<< (xercesc::DOMElement)
          //
          Boolean has_body (has<Traversal::Member> (c) || c.inherits_p ());

          os << "void" << endl
             << "operator<< (::xercesc::DOMElement& e," << endl
             << "const " << type_name << "&" << (has_body ? " i" : "") << ")"
             << "{";

          if (c.inherits_p ())
          {
            os << "e << static_cast< const ";
            inherits (c, inherits_);
            os << "& > (i);"
               << endl;
          }
          else
          {
            // Clear the element.
            //
            os << "while (::xercesc::DOMNode* n = e.getFirstChild ())" << endl
               << "e.removeChild (n);"
               << endl;
          }

          {
            enter_scope (name);

            Traversal::Names names_member;
            Element element (*this, type_name);
            Attribute attribute (*this);

            names_member >> element;
            names_member >> attribute;

            names (c, names_member);

            leave_scope ();
          }

          os << "}";


          Boolean simple (true);
          {
            IsSimpleType t (simple);
            t.dispatch (c);
          }

          if (simple)
          {
            // operator<< (xercesc::DOMAttr)
            //
            os << "void" << endl
               << "operator<< (::xercesc::DOMAttr&" <<
              (has_body ? " a" : "") << "," << endl
               << "const " << type_name << "&" << (has_body ? " i" : "") << ")"
               << "{";

            if (c.inherits_p ())
            {
              os << "a << static_cast< const ";

              inherits (c, inherits_);

              os << "& > (i);";
            }

            os << "}";


            // operator<< (list_stream)
            //
            os << "void" << endl
               << "operator<< (::xsd::cxx::tree::list_stream< " << char_type
               << " >&" << (has_body ? " l" : "") << "," << endl
               << "const " << type_name << "&" << (has_body ? " i" : "") << ")"
               << "{";

            if (c.inherits_p ())
            {
              os << "l << static_cast< const ";

              inherits (c, inherits_);

              os << "& > (i);";
            }

            os << "}";
          }

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () && c.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " << type_name << " >" << endl
               << "_xsd_" << name << "_type_name_init (" << endl
               << L << "\"" << c.name () << "\"," << endl
               << L << "\"" << xml_ns_name (c) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String name;
        String type_name;

        Traversal::Inherits inherits_;
        BaseTypeName base_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element,
                            GlobalElementBase,
                            protected virtual Context
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (!generate_p (e))
            return;

          SemanticGraph::Type& t (e.type ());

          if (!t.named ())
          {
            String const& name (ename (e));
            String type_name (name + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, name, type_name);
            Union union_ (*this, name, type_name);
            Complex complex (*this, name, type_name);
            Enumeration enumeration (*this, name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (e, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }

          if (e.substitutes_p () &&
	      options.value<CLI::generate_polymorphic> ())
          {
            Type& r (e.substitutes ().root ());

            String const& name (ename (e));
            String type_name;

            if (t.named ())
            {
              std::wostringstream o;

              MemberTypeName type (*this, o);
              type.dispatch (t);

              type_name = o.str ();
            }
            else
            {
              type_name = name + L"::" + unclash (name, "type");
            }

            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " << type_name << " >" << endl
               << "_xsd_" << name << "_element_name_init (" << endl
               << L << "\"" << r.name () << "\"," << endl
               << L << "\"" << r.namespace_ ().name () << "\"," << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_ ().name () << "\");"
               << endl
               << endl;
          }
        }
      };


      struct Serializer: Traversal::Element,
                         GlobalElementBase,
                         protected virtual Context
      {
        Serializer (Context& c)
            : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (!doc_root_p (e))
            return;

          String const& name (ename (e));
          String ns (e.namespace_ ().name ());

          // Note that I am using fq-name in function calls because g++ gets
          // confused if the name is 'type'. (see tests/schema/anonymous)
          //

          Boolean fund (false);
          {
            IsFundamentalType t (fund);
            t.dispatch (e.type ());
          }

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          Boolean polymorphic (
            options.value<CLI::generate_polymorphic> () &&
            e.type ().named () &&
            !fund);


          // To an existing DOM instance.
          //
          os << "void" << endl
             << name << " (::xercesc::DOMDocument& d," << endl
             << "const " << type_name (e) << "& s," << endl
             << flags_type << ")"
             << "{"
             << xml_element_type << " e (*d.getDocumentElement ());"
             << endl;

          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_name_map_instance.
            //
            os << "::xsd::cxx::tree::type_name_map< " << char_type
               << " >& tnm (" << endl
               << "::xsd::cxx::tree::type_name_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "tnm.serialize< " << type_name (e) << " > (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "e," << endl
               << "s);";
          }
          else
          {
            os << "if (e.name () == " << L << "\"" << e.name () << "\""
               << " &&" << endl
               << "e.namespace_ () == " << L << "\"" << ns << "\")"
               << "{"
               << "e.dom_element () << s;"
               << "}"
               << "else"
               << "{"
               << "throw ::xsd::cxx::tree::unexpected_element < " <<
              char_type << " > (" << endl
               << "e.name ()," << endl
               << "e.namespace_ ()," << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << ns << "\");"
               << "}";
          }

          os << "}";


          // To a new DOM instance.
          //
          os << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument >" << endl
             << name << " (const " << type_name (e) << "& s," << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >& m," << endl
             << flags_type << " f)"
             << "{"
             << "try"
             << "{";


          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_name_map_instance.
            //
            os << "::xsd::cxx::tree::type_name_map< " << char_type
               << " >& tnm (" << endl
               << "::xsd::cxx::tree::type_name_map_instance< 0, " <<
              char_type << " > ());"
               << endl;
          }

          os << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl;

          if (polymorphic)
          {
            os << "tnm.serialize< " << type_name (e) << " > (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "m," << endl
               << "s," << endl
               << "f));";
          }
          else
          {
            os << "::xsd::cxx::xml::dom::serialize< " <<
              char_type << " > (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << ns << "\"," << endl
               << "m," << endl
               << "f));";
          }

          os << fq_name (e) << " (*d, s, f);"
             << "return d;"
             << "}"
             << "catch (const ::xsd::cxx::xml::dom::mapping< " << char_type << " >& e)"
             << "{"
             << "throw ::xsd::cxx::tree::no_namespace_mapping< " << char_type << " > (e.name ());"
             << "}"
             << "catch (const ::xsd::cxx::xml::dom::xsi_already_in_use&)"
             << "{"
             << "throw ::xsd::cxx::tree::xsi_already_in_use< " <<
	     char_type << " > ();"
             << "}"
             << "}";


          // To XMLFormatTarget.
          //
          os << "void" << endl
             << name << " (::xercesc::XMLFormatTarget& t," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >& m," << endl
             << "const ::std::basic_string< " << char_type << " >& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << endl
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "h.throw_if_failed< ::xsd::cxx::tree::serialization< " <<
            char_type << " > > ();"
             << "}"
             << "}";

          os << "void" << endl
             << name << " (::xercesc::XMLFormatTarget& t," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >& m," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >& h," << endl
             << "const ::std::basic_string< " << char_type << " >& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

          os << "void" << endl
             << name << " (::xercesc::XMLFormatTarget& t," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >& m," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "const ::std::basic_string< " << char_type << " >& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";


          // To std::ostream.
          //
          os << "void" << endl
             << name << " (::std::ostream& o," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >& m," << endl
             << "const ::std::basic_string< " << char_type << " >& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0);"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << endl
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "h.throw_if_failed< ::xsd::cxx::tree::serialization< " <<
            char_type << " > > ();"
             << "}"
             << "}";

          os << "void" << endl
             << name << " (::std::ostream& o," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >& m," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >& h," << endl
             << "const ::std::basic_string< " << char_type << " >& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0);"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

          os << "void" << endl
             << name << " (::std::ostream& o," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const ::xsd::cxx::xml::dom::namespace_infomap< " <<
	    char_type << " >& m," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "const ::std::basic_string< " << char_type << " >& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

        }

      private:
        String
        type_name (Type& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (t.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (t);

            return o.str ();
          }
          else
          {
            String const& name (ename (e));
            return name + L"::" + unclash (name, "type");
          }
        }
      };
    }

    Void
    generate_serialization_source (Context& ctx,
                                   UnsignedLong first,
                                   UnsignedLong last)
    {
      ctx.os << "#include <ostream>" << endl
             << "#include <xsd/cxx/xml/dom/elements.hxx>" << endl
             << "#include <xsd/cxx/xml/dom/serialization.hxx>" << endl
             << "#include <xsd/cxx/tree/error-handler.hxx>" << endl
             << endl;

      if (ctx.options.value<CLI::generate_polymorphic> ())
      {
        ctx.os << "#include <xsd/cxx/tree/type-name-map.hxx>" << endl
               << endl;

        ctx.os << "namespace _xsd"
               << "{"
               << "static" << endl
               << "::xsd::cxx::tree::type_name_plate< 0, " <<
          ctx.char_type << " >" << endl
               << "type_name_plate_init;"
               << "}";
      }

      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names names_ns, names;

      Namespace ns (ctx, false, first, last);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);
      Serializer serializer (ctx);


      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;
      names >> serializer;


      schema.dispatch (ctx.schema_root);
    }
  }
}
