// file      : xsd/cxx/tree/tree-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/tree-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cult/containers/list.hxx>

#include <iostream>

using std::wcerr;

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
          {
            name = ename (l);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (l, name) && !name)
              return;
          }


          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          if (!options.value<CLI::suppress_parsing> ())
          {
            String base (L"::xsd::cxx::tree::list< " +
                         item_type_name (l) +
                         L", " + char_type + L" >");

            // c-tor (xercesc::DOMElement)
            //
            os << scoped_name << "::" << endl
               << name << " (const ::xercesc::DOMElement& e," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << any_simple_type << " (e, f, c)," << endl
               << base << " (e, f, c)"
               << "{"
               << "}";

            // c-tor (xercesc::DOMAttr)
            //
            os << scoped_name << "::" << endl
               << name << " (const ::xercesc::DOMAttr& a," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << any_simple_type << " (a, f, c)," << endl
               << base << " (a, f, c)"
               << "{"
               << "}";

            // c-tor (string const&, xercesc::DOMElement)
            //
            os << scoped_name << "::" << endl
               << name << " (const " << string_type << "& s," << endl
               << "const ::xercesc::DOMElement* e," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << any_simple_type << " (s, e, f, c)," << endl
               << base << " (s, e, f, c)"
               << "{"
               << "}";
          }

          // _clone
          //
          os << scoped_name << "* " << scoped_name << "::" << endl
             << "_clone (" << flags_type << " f," << endl
             << any_type << "* c) const"
             << "{"
             << "return new " << name << " (*this, f, c);"
             << "}";

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () && l.named ())
          {
            // Note that we are using the original type name.
            //
            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << ename (l) << " >" << endl
               << "_xsd_" << ename (l) << "_type_factory_init (" << endl
               << L << "\"" << l.name () << " " << xml_ns_name (l) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String
        item_type_name (Type& l)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (l.argumented ().type ());

          return o.str ();
        }

        /*
        String
        type_name (Type& l)
        {
          if (l.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (l);

            return o.str ();
          }
          else
          {
            enter_scope (name);
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }
        */

      private:
        String name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c, String const& name_ = L"")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
          {
            name = ename (u);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (u, name) && !name)
              return;
          }


          String base (xs_ns_name () + L"::string");

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          if (!options.value<CLI::suppress_parsing> ())
          {
            // c-tor (xercesc::DOMElement)
            //
            os << scoped_name << "::" << endl
               << name << " (const ::xercesc::DOMElement& e," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << base << " (e, f, c)"
               << "{"
               << "}";

            // c-tor (xercesc::DOMAttr)
            //
            os << scoped_name << "::" << endl
               << name << " (const ::xercesc::DOMAttr& a," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << base << " (a, f, c)"
               << "{"
               << "}";

            // c-tor (string const&, xercesc::DOMElement)
            //
            os << scoped_name << "::" << endl
               << name << " (const " << string_type << "& s," << endl
               << "const ::xercesc::DOMElement* e," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << base << " (s, e, f, c)"
               << "{"
               << "}";
          }

          // _clone
          //
          os << scoped_name << "* " << scoped_name << "::" << endl
             << "_clone (" << flags_type << " f," << endl
             << any_type << "* c) const"
             << "{"
             << "return new " << name << " (*this, f, c);"
             << "}";

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () && u.named ())
          {
            // Note that we are using the original type name.
            //
            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << ename (u) << " >" << endl
               << "_xsd_" << ename (u) << "_type_factory_init (" << endl
               << L << "\"" << u.name () << " " << xml_ns_name (u) << "\");"
               << endl
               << endl;
          }
        }

      private:
        /*
        String
        type_name (Type& u)
        {
          if (u.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (u);

            return o.str ();
          }
          else
          {
            enter_scope (name);
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }
        */

      private:
        String name;
      };


      // Enumeration mapping.
      //

      struct EnumeratorLiteral: Traversal::Enumerator,
                                protected virtual Context
      {
        EnumeratorLiteral (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          os << L << "\"" << e.name () << "\"";
        }
      };


      //
      //
      struct LiteralInfo
      {
        LiteralInfo (String const& value, String const& name)
            : value_ (value), name_ (name)
        {
        }

        String value_;
        String name_;
      };

      Boolean
      operator< (LiteralInfo const& x, LiteralInfo const& y)
      {
        return x.value_ < y.value_;
      }

      typedef Cult::Containers::List<LiteralInfo> LiteralInfoList;


      // Populate LiteralInfoList
      //
      struct EnumeratorLiteralInfo: Traversal::Enumerator,
                                    protected virtual Context

      {
        EnumeratorLiteralInfo (Context& c, LiteralInfoList& list)
            : Context (c), list_ (list)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          list_.push_back (LiteralInfo (e.name (), ename (e)));
        }

      private:
        LiteralInfoList& list_;
      };


      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c, String name_ = L"")
            : Context (c),
              name (name_),
              enumerator_literal_ (c)
        {
          names_enumerator_literal_ >> enumerator_literal_;
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
          {
            name = ename (e);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (e, name) && !name)
              return;
          }


          Boolean string_based (false);
          {
            IsStringBasedType t (string_based);
            t.dispatch (e);
          }

          SemanticGraph::Enumeration* be (0);
          Boolean enum_based (false);
          if (string_based)
          {
            IsEnumBasedType t (be);
            t.dispatch (e);

            enum_based = (be != 0);
          }

          UnsignedLong enum_count (0);

          for (Type::NamesIterator i (e.names_begin ()), end (e.names_end ());
               i != end; ++i)
            ++enum_count;

          String base; // base type name
          {
            std::wostringstream o;

            BaseTypeName base_type (*this, o);
            Traversal::Inherits inherits_type (base_type);

            inherits (e, inherits_type);
            base = o.str ();
          }

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          if (!options.value<CLI::suppress_parsing> ())
          {
            // c-tor (xercesc::DOMElement)
            //
            os << scoped_name << "::" << endl
               << name << " (const ::xercesc::DOMElement& e," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << base << " (e, f, c)"
               << "{";

            if (string_based)
              os << "_xsd_" << name << "_convert ();";

            os << "}";


            // c-tor (xercesc::DOMAttr)
            //
            os << scoped_name << "::" << endl
               << name << " (const ::xercesc::DOMAttr& a," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << base << " (a, f, c)"
               << "{";

            if (string_based)
            {
              os << "_xsd_" << name << "_convert ();";
            }

            os << "}";


            // c-tor (string const&, xercesc::DOMElement)
            //
            os << scoped_name << "::" << endl
               << name << " (const " << string_type << "& s," << endl
               << "const ::xercesc::DOMElement* e," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << base << " (s, e, f, c)"
               << "{";

            if (string_based)
            {
              os << "_xsd_" << name << "_convert ();";
            }

            os << "}";
          }

          // _clone
          //
          os << scoped_name << "* " << scoped_name << "::" << endl
             << "_clone (" << flags_type << " f," << endl
             << any_type << "* c) const"
             << "{"
             << "return new " << name << " (*this, f, c);"
             << "}";

          // convert
          //
          // @@ TODO: expected list
          //
          if (string_based)
          {
            String i_name (L"_xsd_" + name + L"_indexes_");

            os << scoped_name /*tn*/ << "::_xsd_" << name << " " <<
              scoped_name << "::" << endl
               << "_xsd_" << name << "_convert () const"
               << "{"
               << "::xsd::cxx::tree::enum_comparator< " << char_type <<
              " > c (_xsd_" << name << "_literals_);"
               << "const _xsd_" << name << "* i (::std::lower_bound (" << endl
               << i_name << "," << endl
               << i_name << " + " << enum_count << "," << endl
               << "*this," << endl
               << "c));"
               << endl
               << "if (i == " << i_name << " + " << enum_count << " || " <<
              "_xsd_" << name << "_literals_[*i] != *this)"
               << "{"
               << "throw ::xsd::cxx::tree::unexpected_enumerator < " <<
              char_type << " > (*this);"
               << "}"
               << "return *i;"
               << "}";
          }

          // literals and indexes
          //
          if (string_based)
          {
            if (enum_based)
            {
              os << "const " << char_type << "* const* " << scoped_name << "::" << endl
                 << "_xsd_" << name << "_literals_ = " <<
                fq_name (*be) << "::_xsd_" << ename (*be) << "_literals_;"
                 << endl;
            }
            else
            {
              os << "const " << char_type << "* const " << scoped_name << "::" << endl
                 << "_xsd_" << name << "_literals_[" << enum_count << "] ="
                 << "{";

              {
                enter_scope (name);
                names<Enumeration> (
                  e, names_enumerator_literal_, 0, 0, 0, &Enumeration::comma);
                leave_scope ();
              }

              os << "};";
            }


            LiteralInfoList l;
            {
              enter_scope (name);

              EnumeratorLiteralInfo enumerator (*this, l);
              Traversal::Names names_enumerator (enumerator);

              names (e, names_enumerator);

              leave_scope ();
            }

            l.sort ();

            os << "const " << scoped_name << "::_xsd_" << name << " " <<
              scoped_name << "::" << endl
               << "_xsd_" << name << "_indexes_[" << enum_count << "] ="
               << "{";

            String fq_name (ns_scope + L"::" + scoped_name);

            for (LiteralInfoList::Iterator
                   b (l.begin ()), i (b), end (l.end ()); i != end; ++i)
            {
              if (i != b)
                os << "," << endl;

              os << fq_name << "::" << i->name_;
            }

            os << "};";
          }

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () && e.named ())
          {
            // Note that we are using the original type name.
            //
            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << ename (e) << " >" << endl
               << "_xsd_" << ename (e) << "_type_factory_init (" << endl
               << L << "\"" << e.name () << " " << xml_ns_name (e) << "\");"
               << endl
               << endl;
          }
        }

        virtual Void
        comma (Type&)
        {
          os << "," << endl;
        }

      private:
        /*
        String
        type_name (Type& e)
        {
          if (e.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (e);

            return o.str ();
          }
          else
          {
            enter_scope (name);
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }
        */

      private:
        String name;

      private:
        Traversal::Names names_enumerator_literal_;
        EnumeratorLiteral enumerator_literal_;

      };


      // For anonymous types.
      //
      //
      template <typename Complex>
      struct Member : Traversal::Member,
                      protected virtual Context
      {
        Member (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            Traversal::Belongs belongs_;

            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }

          // default_value
          //
          if (m.default_ ())
          {
            Boolean simple (true);

            if (m.is_a<SemanticGraph::Element> ())
            {
              IsSimpleType test (simple);
              test.dispatch (t);
            }

            if (simple)
            {
              Boolean fund (false);
              {
                IsFundamentalType t (fund);
                t.dispatch (m.type ());
              }

              os << "// " << scope << "::" << name << endl
                 << "// " << endl
                 << endl;

              String type (iscope_name (name, "type"));
              String q_type (scope + L"::" + type);

              if (fund)
              {
                String ftype;
                {
                  std::wostringstream o;
                  MemberTypeName t (*this, o);
                  t.dispatch (m.type ());
                  ftype = o.str ();
                }

                os << "const " << q_type << " " << scope << "::" << name <<
                  "::" << unclash (name, "default_value") << "_ (" << endl
                   << "::xsd::cxx::tree::traits< " << ftype << ", " <<
                  char_type << " >::create (" << endl
                   << string_type << " (" << L << "\"" << m.value () <<
                  "\"), 0, 0, 0));"
                   << endl;
              }
              else
              {
                os << "const " << q_type << " " << scope << "::" << name <<
                  "::" << unclash (name, "default_value") << "_ (" << endl
                   << string_type << " (" << L << "\"" << m.value () <<
                  "\"), 0, 0, 0);"
                   << endl;
              }
            }
          }
        }
      };


      struct Element : Traversal::Element, protected virtual Context
      {
        Element (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (skip (e)) return;

          String const& name (ename (e));
          String tr (iscope_name (name, "traits")); // traits type name
          String type (iscope_name (name, "type"));

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          SemanticGraph::Type& t (e.type ());


          Boolean fund (false);
          {
            IsFundamentalType traverser (fund);
            traverser.dispatch (t);
          }

          Boolean polymorphic (!fund
                               && options.value<CLI::generate_polymorphic> ()
                               && t.named ());

          os << "// " << name << endl
             << "//" << endl
             << "{";

          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_factory_map_instance.
            //
            os << "::xsd::cxx::tree::type_factory_map< " << char_type <<
              " >& tfm (" << endl
               << "::xsd::cxx::tree::type_factory_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "::std::auto_ptr< " << type << " > r (" << endl
               << "tfm.create< " << type << " > (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" <<
              (e.qualified () ? e.namespace_ ().name () : L"") << "\"," << endl
               << (e.global () ? "true" : "false") << "," << endl
               << (e.qualified () ? "true" : "false") << "," << endl
               << "e," << endl
               << "f | " << flags_type << "::not_root," << endl
               << "this));"
               << endl
               << "if (r.get () != 0)"
               << "{";
          }
          else
          {
            if (e.qualified ())
            {
              os << "if (e.name () == " << L << "\"" << e.name () << "\" && " <<
                "e.namespace_ () == " << L << "\"" <<
                e.namespace_ ().name () << "\")"
                 << "{";
            }
            else
            {
              os << "if (e.name () == " << L << "\"" << e.name () << "\" && " <<
                "e.namespace_ ().empty ())"
                 << "{";
            }

            if (!fund)
            {
              os << "::std::auto_ptr< " << type << " > r (" << endl
                 << tr << "::create (" << endl
                 << "e.dom_element ()," << endl
                 << "f | " << flags_type << "::not_root," << endl
                 << "this));"
                 << endl;
            }
          }


          // Checks. Disabled at the moment since they make it impossible to
          // parse valid instances where the same element is used in both
          // base and derived types. See the cxx/tree/name-clash/inheritance
          // test for details.
          //
          //
          if (max (e) != 1)
          {
            // sequence
            //
          }
          else if (min (e) == 0)
          {
            // optional
            //
            os << "if (this->" << name << " ())" << endl
               << "continue;";

              /*
               << "{"
               << "throw ::xsd::cxx::tree::unexpected_element< " <<
              char_type << " > (" << endl
               << "e.name (), e.namespace_ ()," << endl
               << L << "\"\", " << L << "\"\");"
               << "}";
              */
          }
          else
          {
            // one
            //
            os << "if (_xsd_" << name << "_.present ())" << endl
               << "continue;";

              /*
               << "{"
               << "throw ::xsd::cxx::tree::unexpected_element< " <<
              char_type << " > (" << endl
               << "e.name (), e.namespace_ ()," << endl
               << L << "\"\", " << L << "\"\");"
               << "}";
              */
          }


          if (polymorphic || !fund)
          {
            if (max (e) != 1)
            {
              // sequence
              //
              os << "this->" << name << " ().push_back (r);";
            }
            else if (min (e) == 0)
            {
              // optional
              //
              os << "this->" << name << " (r);";
            }
            else
            {
              // one
              //
              os << "this->" << name << " (r);";
            }
          }
          else
          {
            if (max (e) != 1)
            {
              // sequence
              //
              os << "this->" << name << " ().push_back (" << endl
                 << tr << "::create (" << endl
                 << "e.dom_element ()," << endl
                 << "f | " << flags_type << "::not_root," << endl
                 << "this));";
            }
            else if (min (e) == 0)
            {
              // optional
              //
              os << "this->" << name << " (" << endl
                 << tr << "::create (" << endl
                 << "e.dom_element ()," << endl
                 << "f | " << flags_type << "::not_root," << endl
                 << "this));";
            }
            else
            {
              // one
              //
              os << "this->" << name << " (" << endl
                 << tr << "::create (" << endl
                 << "e.dom_element ()," << endl
                 << "f | " << flags_type << "::not_root," << endl
                 << "this));";
            }
          }

          os << "continue;"
             << "}" // if block
             << "}";
        }
      };

      struct ElementTest : Traversal::Element, protected virtual Context
      {
        ElementTest (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (skip (e)) return;

          String const& name (ename (e));

          if (max (e) == 1 && min (e) == 1)
          {
            // one
            //
            os << "if (!_xsd_" << name << "_.present ())"
               << "{"
               << "throw ::xsd::cxx::tree::expected_element< " <<
              char_type << " > (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" <<
              (e.qualified () ? e.namespace_ ().name () : L"") << "\");"
               << "}";
          }
        }
      };

      struct Attribute : Traversal::Attribute, protected virtual Context
      {
        Attribute (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));
          String tr (iscope_name (name, "traits")); // traits type name

          if (a.qualified ())
          {
            os << "if (a.name () == " << L << "\"" << a.name () << "\" && " <<
              "a.namespace_ () == " << L << "\"" <<
              a.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (a.name () == " << L << "\"" << a.name () << "\" && " <<
              "a.namespace_ ().empty ())"
               << "{";
          }

          Boolean fund (false);

          {
            IsFundamentalType traverser (fund);
            traverser.dispatch (a.type ());
          }

          if (fund)
          {
            os << "this->" << name << " (" << endl
               << tr << "::create (" << endl
               << "a.dom_attribute ()," << endl
               << "f | " << flags_type << "::not_root," << endl
               << "this));";
          }
          else
          {
            String type (iscope_name (name, "type"));

            os << "::std::auto_ptr< " << type << " > r (" << endl
               << tr << "::create (" << endl
               << "a.dom_attribute ()," << endl
               << "f | " << flags_type << "::not_root," << endl
               << "this));"
               << endl
               << "this->" << name << " (r);";
          }

          os << "continue;"
             << "}";

        }
      };


      struct AttributeTest : Traversal::Attribute, protected virtual Context
      {
        AttributeTest (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));

          if (!a.optional () || a.default_ ())
          {
            os << "if (!_xsd_" << name << "_.present ())"
               << "{";

            if (a.default_ ())
              os << "this->" << name << " (" << endl
                 << iscope_name (name, "traits") << "::create (" << endl
                 << string_type << " (" << L << "\"" << a.value () <<
                "\")," << endl
                 << "&e," << endl
                 << "f | " << flags_type << "::not_root," << endl
                 << "this));";
            else
              os << "throw ::xsd::cxx::tree::expected_attribute< " <<
                char_type << " > (" << endl
                 << L << "\"" << a.name () << "\"," << endl
                 << L << "\"" <<
                (a.qualified () ? a.namespace_ ().name () : L"") << "\");";

            os << "}";
          }
        }
      };


      //
      //
      struct CtorBase : Traversal::Complex,
                        Traversal::Enumeration,
                        Traversal::Type,
                        protected virtual Context
      {
        CtorBase (Context& c, String const& name_)
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          os << name;
        }

        virtual Void
        traverse (SemanticGraph::Enumeration& e)
        {
          os << name;
        }

        Void
        traverse (SemanticGraph::Complex& c)
        {
          Args args (*this, name);
          args.traverse (c);
        }

      private:
        struct Args : Traversal::Complex,
                      Traversal::Enumeration,
                      Traversal::Type,
                      Traversal::Member,
                      protected virtual Context
        {
          Args (Context& c, String const& name_)
              : Context (c), name (name_), first_ (true)
          {
            *this >> inherits_ >> *this;
            *this >> names_ >> *this;
          }

          virtual Void
          traverse (SemanticGraph::Type& t)
          {
            os << comma () << name;
          }

          virtual Void
          traverse (SemanticGraph::Enumeration& e)
          {
            os << comma () << name;
          }

          virtual Void
          traverse (SemanticGraph::Member& m)
          {
            if (skip (m)) return;

            // Note that we are not including attributes with default
            // values here.
            if (min (m) == 1 && max (m) == 1)
            {
              // one
              //
              os << comma () << "_xsd_" << ename (m);
            }
          }

          using Complex::traverse;

        private:
          String
          comma ()
          {
            Boolean tmp (first_);
            first_ = false;
            return tmp ? "" : ",\n";
          }

        private:
          String name;
          Boolean first_;

          Traversal::Inherits inherits_;
          Traversal::Names names_;
        };

        String name;
      };


      struct CtorMember : Traversal::Member,
                          protected virtual Context
      {
        CtorMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          if (max (m) != 1)
          {
            // sequence
            //
            os << "," << endl
               << "_xsd_" << name << "_ (" << flags_type << " (), this)";
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            os << "," << endl
               << "_xsd_" << name << "_ (" << flags_type << " (), this)";
          }
          else
          {
            // one
            //

            if (min (m) == 0)
            {
              // This is an attribute with default value. We are going
              // to initialize it to its default value.
              //
              os << "," << endl
                 << "_xsd_" << name << "_ (" << name << "::" <<
                unclash (name, "default_value") << " ()," << endl
                 << flags_type << " ()," << endl
                 << "this)";
            }
            else
            {
              os << "," << endl
                 << "_xsd_" << name << "_ (_xsd_" << name << "," << endl
                 << flags_type << " ()," << endl
                 << "this)";
            }
          }
        }
      };


      struct CopyMember : Traversal::Member,
                          protected virtual Context

      {
        CopyMember (Context& c, String const& arg_name_)
            : Context (c), arg_name (arg_name_)
        {
        }

        virtual Void
        traverse (SemanticGraph::Member& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          /*
          if (max (m) != 1)
          {
            // sequence
            //
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
          }
          else
          {
            //
            // one
          }
          */

          os << "," << endl
             << "_xsd_" << name << "_ (" <<
            arg_name << "._xsd_" << name << "_," << endl
             << "f | " << flags_type << "::not_root," << endl
             << "this)";
        }

      private:
        String arg_name;
      };


      // C-tor initializer.
      //
      struct ElementCtorMember : Traversal::Member,
                                 protected virtual Context
      {
        ElementCtorMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          /*
          if (max (m) != 1)
          {
            // sequence
            //
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
          }
          else
          {
            // one
            //
          }
          */

          os << "," << endl
             << "_xsd_" << name << "_ (" <<
            "f | " << flags_type << "::not_root, this)";
        }
      };


      // Default c-tor member initializer.
      //
      struct DefaultCtorMemberInit : Traversal::Member,
                                     protected virtual Context
      {
        DefaultCtorMemberInit (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          os << "," << endl
             << "_xsd_" << name << "_ (" << flags_type << " (), this)";
        }
      };

      //
      //
      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              member_ (c),
              element_ (c),
              element_test_ (c),
              attribute_ (c),
              attribute_test_ (c),
              default_ctor_member_init_ (c),
              from_base_ctor_arg_ (c, true),
              ctor_member_ (c),
              element_ctor_member_ (c)
        {
          names_member_ >> member_;

          names_element_ >> element_;
          names_element_test_ >> element_test_;

          names_attribute_ >> attribute_;
          names_attribute_test_ >> attribute_test_;

          default_ctor_member_init_names_ >> default_ctor_member_init_;

          from_base_ctor_arg_names_ >> from_base_ctor_arg_;

          ctor_member_names_ >> ctor_member_;

          element_ctor_member_names_ >> element_ctor_member_;
        }


        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
          {
            name = ename (c);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (c, name) && !name)
              return;
          }

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          String base; // base type name
          if (c.inherits_p ())
          {
            std::wostringstream o;

            BaseTypeName base_type (*this, o);
            Traversal::Inherits inherits_type (base_type);

            // Cannot use inherits_none here because need the
            // result in a string.
            //
            inherits (c, inherits_type);

            base = o.str ();
          }
          else
            base = any_type;

          os << "// " << scoped_name << endl
             << "//" << endl
             << endl;

          // Go after anonymous types first.
          //
          {
            enter_scope (name);
            names (c, names_member_);
            leave_scope ();
          }

          // default c-tor
          //
          if (options.value<CLI::generate_default_ctor> ())
          {
            Boolean generate (false);
            {
              GenerateDefaultCtor t (*this, generate);
              t.traverse (c);
            }

            if (generate)
            {
              os << scoped_name << "::" << endl
                 << name << " ()" << endl
                 << ": " << base << " ()";

              {
                enter_scope (name);
                names (c, default_ctor_member_init_names_);
                leave_scope ();
              }

              os << "{"
                 << "}";
            }
          }

          // c-tor (base, all-non-optional-members)
          //
          if (options.value<CLI::generate_from_base_ctor> ())
          {
            Boolean generate (false);
            {
              GenerateFromBaseCtor t (*this, generate);
              t.traverse (c);
            }

            if (generate)
            {
              os << scoped_name << "::" << endl
                 << name << " (const _xsd_" << name << "::base_& _xsd_" << name;

              enter_scope (name);
              names (c, from_base_ctor_arg_names_);
              leave_scope ();

              os << ")" << endl
                 << ": " << base << " (_xsd_" << name << ")";

              enter_scope (name);
              names (c, ctor_member_names_);
              leave_scope ();

              os << "{"
                 << "}";
            }
          }

          // c-tor (all-non-optional-members)
          //
          os << scoped_name << "::" << endl
             << name << " (";

          {
            enter_scope (name);

            CtorArgs ctor_args (*this, L"_xsd_" + name, true);
            ctor_args.dispatch (c);

            leave_scope ();
          }

          os << ")" << endl
             << ": " << base << " (";

          {
            enter_scope (name);

            CtorBase base (*this, L"_xsd_" + name);
	    Traversal::Inherits inherits_base (base);

            inherits (c, inherits_base);

            leave_scope ();
          }

          os << ")";

          {
            enter_scope (name);
            names (c, ctor_member_names_);
            leave_scope ();
          }

          os << "{"
             << "}";


          // copy c-tor
          //
          os << scoped_name << "::" << endl
             << name << " (const " << name << "& _xsd_" << name << "," << endl
             << flags_type << " f," << endl
             << any_type << "* c)" << endl
             << ": " << base << " (_xsd_" << name << ", f, c)";

          {
            enter_scope (name);

            CopyMember copy_member (*this, L"_xsd_" + name);
            Traversal::Names names_member (copy_member);

            names<Complex> (c, names_member);

            leave_scope ();
          }

          os << "{"
             << "}";

          if (!options.value<CLI::suppress_parsing> ())
          {
            //
            //
            Boolean he (has<Traversal::Element> (c));
            Boolean ha (has<Traversal::Attribute> (c));

            // c-tor (xercesc::DOMElement)
            //
            os << scoped_name << "::" << endl
               << name << " (const ::xercesc::DOMElement& e," << endl
               << flags_type << " f," << endl
               << any_type << "* c)" << endl
               << ": " << base << " (e, f, c)";

            {
              enter_scope (name);
              names (c, element_ctor_member_names_);
              leave_scope ();
            }

            os << "{";


            if (he || ha)
            {
              os << unclash (name, "parse") << " (e, f);";
            }

            os << "}";

            Boolean simple (true);
            {
              IsSimpleType t (simple);
              t.dispatch (c);
            }

            if (simple)
            {
              // c-tor (xercesc::DOMAttr)
              //
              os << scoped_name << "::" << endl
                 << name << " (const ::xercesc::DOMAttr& a," << endl
                 << flags_type << " f," << endl
                 << any_type << "* c)" << endl
                 << ": " << base << " (a, f, c)"
                 << "{"
                 << "}";

              // c-tor (string const&, xercesc::DOMElement)
              //
              os << scoped_name << "::" << endl
                 << name << " (const " << string_type << "& s," << endl
                 << "const ::xercesc::DOMElement* e," << endl
                 << flags_type << " f," << endl
                 << any_type << "* c)" << endl
                 << ": " << base << " (s, e, f, c)"
                 << "{"
                 << "}";
            }

            if (he || ha)
            {
              os << "void " << scoped_name << "::" << endl
                 << unclash (name, "parse") << " (" <<
                "const ::xercesc::DOMElement& e, " << flags_type << " f)"
                 << "{"
                 << parser_type << " p (e);"
                 << endl;

              if (he)
              {
                os << "while (p.more_elements ())"
                   << "{"
                   << "const " << xml_element_type << " e (p.next_element ());"
                   << endl;

                {
                  enter_scope (name);

                  names (c, names_element_);

                  leave_scope ();
                }

                //@@ TODO: expected list
                //
                // @@
                // This doesn't play well with inheritance because we
                // don't expect base's elements.
                //
                //<< "throw ::xsd::cxx::tree::unexpected_element < " <<
                //char_type << " > (n, " << L << "\"" << "\");"

                os << "}"; // while loop

                // Make sure all non-optional elements are set.
                //
                {
                  enter_scope (name);
                  names (c, names_element_test_);
                  leave_scope ();
                }
              }

              if (ha)
              {
                os << "while (p.more_attributes ())"
                   << "{"
                   << "const " << xml_attribute_type <<
		  " a (p.next_attribute ());"
                   << endl;

                {
                  enter_scope (name);
                  names (c, names_attribute_);
                  leave_scope ();
                }

                // os << "{"  // else block
                // @@
                // This doesn't play well with inheritance because we
                // don't expect base's attributes. Also there are other
                // "special" attributes such as xmlns, etc.
                //
                // << "throw ::xsd::cxx::tree::unexpected_attribute < " <<
                // char_type << " > (n, " << L << "\"" << "\");"
                // << "}";
                os << "}"; // while loop

                // Make sure all non-optional attributes are set.
                //
                {
                  enter_scope (name);
                  names (c, names_attribute_test_);
                  leave_scope ();
                }
              }

              os << "}";
            }
          }

          // _clone
          //
          // Was: type_name (c) instead of scoped_name
          //
          os << scoped_name << "* " << scoped_name << "::" << endl
             << "_clone (" << flags_type << " f," << endl
             << any_type << "* c) const"
             << "{"
             << "return new " << name << " (*this, f, c);"
             << "}";


          // Register with type factory map.
          //
          // Was: type_name (c) instead of scoped_name
          //
          if (options.value<CLI::generate_polymorphic> () && c.named ())
          {
            // Note that we are using the original type name.
            //
            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << ename (c) << " >" << endl
               << "_xsd_" << ename (c) << "_type_factory_init (" << endl
               << L << "\"" << c.name () << " " << xml_ns_name (c) << "\");"
               << endl
               << endl;
          }
        }

      private:
        /*
        String
        type_name (Type& c)
        {
          if (c.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (c);

            return o.str ();
          }
          else
          {
            enter_scope (name);
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }
        */

      private:
        String name;

        Member<Complex> member_;
        Traversal::Names names_member_;

        Traversal::Names names_element_;
        Traversal::Names names_element_test_;

        Traversal::Names names_attribute_;
        Traversal::Names names_attribute_test_;

        Element element_;
        ElementTest element_test_;

        Attribute attribute_;
        AttributeTest attribute_test_;

        DefaultCtorMemberInit default_ctor_member_init_;
        Traversal::Names default_ctor_member_init_names_;

        FromBaseCtorArg from_base_ctor_arg_;
        Traversal::Names from_base_ctor_arg_names_;

        CtorMember ctor_member_;
        Traversal::Names ctor_member_names_;

        ElementCtorMember element_ctor_member_;
        Traversal::Names element_ctor_member_names_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element,
                            GlobalElementBase,
                            protected virtual Context
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (!generate_p (e))
            return;

          SemanticGraph::Type& t (e.type ());

          if (!t.named ())
          {
            String const& name (ename (e));

            Traversal::Belongs belongs_;

            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (e, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }

          if (e.substitutes_p () &&
              options.value<CLI::generate_polymorphic> ())
          {
            Type& r (e.substitutes ().root ());

            String const& name (ename (e));
            String type_name;

            if (t.named ())
            {
              std::wostringstream o;

              MemberTypeName type (*this, o);
              type.dispatch (t);

              type_name = o.str ();
            }
            else
            {
              type_name = name + L"::" +
                (name == L"type" ? L"type_" : L"type");
            }

            os << "static" << endl
               << "::xsd::cxx::tree::type_factory_initializer< 0, " <<
              char_type << ", " << type_name << " >" << endl
               << "_xsd_" << name << "_element_factory_init (" << endl
               << L << "\"" << r.name () << " " <<
              r.namespace_ ().name () << "\", " << endl
               << L << "\"" << e.name () << " " <<
              e.namespace_ ().name () << "\");"
               << endl
               << endl;
          }
        }
      };
    }

    Void
    generate_tree_source (Context& ctx,
                          UnsignedLong first,
                          UnsignedLong last)
    {
      if (!ctx.options.value<CLI::suppress_parsing> ())
        ctx.os << "#include <xsd/cxx/xml/dom/elements.hxx>" << endl
               << "#include <xsd/cxx/xml/dom/parser.hxx>" << endl
               << endl;

      if (ctx.options.value<CLI::generate_polymorphic> ())
      {
        ctx.os << "#include <xsd/cxx/tree/type-factory-map.hxx>" << endl
               << endl;

        ctx.os << "namespace _xsd"
               << "{"
               << "static" << endl
               << "::xsd::cxx::tree::type_factory_plate< 0, " <<
          ctx.char_type << " >" << endl
               << "type_factory_plate_init;"
               << "}";
      }

      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx, false, first, last);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;

      schema.dispatch (ctx.schema_root);
    }
  }
}
